package ga.problems;

import ga.Algorithm;
import ga.Individual;
import ga.chromosomes.Chromosome;
import ga.chromosomes.IntChromosome;
import ga.problems.helpers.City;
import ga.problems.helpers.UniqueIntegersChromosomeStrategy;

import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.Map;

public class ProblemTSP extends Problem {
    private Map<Integer, City> cities = new HashMap<Integer, City>();
    private double[][] weightCache; // for performance speedup
    private static final double WEIGHT_INIT_VALUE = -1.0;

    public ProblemTSP(double crossoverProbability, double mutationsPerChromosome, String fileName) {
        readCitiesFromFile(fileName);
        allelesPerChromosome = cities.size();
        chromosomeStrategy = new UniqueIntegersChromosomeStrategy(
            allelesPerChromosome,
            crossoverProbability,
            mutationsPerChromosome,
            this);
        initializeWeightCache();
    }

    private void initializeWeightCache() {
        weightCache = new double[allelesPerChromosome][allelesPerChromosome];
        for (int i = 0; i < allelesPerChromosome; i++)
            weightCache[0][i] = WEIGHT_INIT_VALUE;
        for (int i = 1; i < allelesPerChromosome; i++)
            System.arraycopy(weightCache[0], 0, weightCache[i], 0, allelesPerChromosome);
    }

    /**
     * Get a template chromosome, according to the Zero-City Anchor concept
     */
    public Chromosome getTemplateChromosome() {
        IntChromosome result = (IntChromosome) super.getTemplateChromosome();
        // we want all searches to be "anchored" by the first city.
        // Since all routes are cyclical, there are equivalence classes in the
        // solution space, each consisting of isomorphic graphs. We want to
        // evaluate all members of a given equivalence class as one single
        // entity.
        // Practically, this means that our crossover mechanism will always
        // create a new route still anchored by the first city.
        int swapIndex = result.getAlleleIndexByValue(0);
        int swapCityId = result.getAllele(0);
        result.setAllele(0, 0); // "anchor" city is cityID=0 at position=0
        result.setAllele(swapIndex, swapCityId);
        return result;
    }

    /**
     * fitness score of an individual is defined as the total cost (weight) of
     * the route represented by that individual.
     */
    public double evaluate(Individual individual) {
        double routeWeight = 0.0;
        IntChromosome chromosome = (IntChromosome) individual.getChromosome();
        for (int i = 1; i < allelesPerChromosome; i++) {
            int cityId1 = chromosome.getAllele(i - 1);
            int cityId2 = chromosome.getAllele(i);
            if (weightCache[cityId1][cityId2] == WEIGHT_INIT_VALUE) {
                // cache miss
                weightCache[cityId1][cityId2] = cities.get(cityId1).distanceFrom(
                    cities.get(cityId2));
                weightCache[cityId2][cityId1] = weightCache[cityId1][cityId2];
            }
            routeWeight += weightCache[cityId1][cityId2];
        }
        individual.setScore(routeWeight);
        return routeWeight;
    }

    public boolean resultHasBeenFound(Algorithm geneticAlgorithm, double targetScore) {
        return geneticAlgorithm.getBestKnownIndividual().getScore() <= targetScore;
    }

    public void readCitiesFromFile(String fileName) {

        FileInputStream fileInputStream;
        try {
            fileInputStream = new FileInputStream(fileName);
            DataInputStream inputStream = new DataInputStream(fileInputStream);
            BufferedReader reader = new BufferedReader(new InputStreamReader(inputStream));
            String fileLine;
            int i = 0;
            while ((fileLine = reader.readLine()) != null) {
                String[] latLong = fileLine.split(" ");
                cities.put(i++, new City(
                    Double.parseDouble(latLong[0]),
                    Double.parseDouble(latLong[1])));
            }
            inputStream.close();
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }

    public int isScore1BetterThanScore2(double score1, double score2) {
        return compareScoresForMinimization(score1, score2);
    }
}
