package ga.problems.helpers;

import ga.chromosomes.Chromosome;
import ga.chromosomes.IntChromosome;
import ga.problems.Problem;

public class UniqueIntegersChromosomeStrategy extends ChromosomeStrategy {
    public UniqueIntegersChromosomeStrategy(int allelesPerChromosome, double crossoverProbability,
            double mutationsPerChromosome, Problem problemBeingSolved) {
        super(
            allelesPerChromosome,
            crossoverProbability,
            mutationsPerChromosome,
            problemBeingSolved);
    }

    public Chromosome getTemplateChromosome() {
        int[] alleles = getRandomizedListOfSequentialIntegers();
        return new IntChromosome(alleles);
    }

    private int[] getRandomizedListOfSequentialIntegers() {
        RandomIntSequence randomInt = new RandomIntSequence(allelesPerChromosome);
        int[] result = new int[allelesPerChromosome];
        for (int i = 0; i < allelesPerChromosome; i++) {
            result[i] = randomInt.getNext();
        }
        return result;
    }

    /**
     * Because this Chromosome is specifically designed for the TSP problem,
     * mutations are done by swapping "cities" (alleles), so that every city is
     * always visited exactly once (after any number of mutations)
     */
    public void mutate(Chromosome chromosome) {
        for (int i = 0; i < chromosome.getLength(); i++)
            if (random.nextDouble() <= mutationProbability) {
                // mutate by swapping
                int j = random.nextInt(chromosome.getLength());
                int oldI = ((IntChromosome) chromosome).getAllele(i);
                ((IntChromosome) chromosome).setAllele(i, ((IntChromosome) chromosome).getAllele(j));
                ((IntChromosome) chromosome).setAllele(j, oldI);
            }
    }

    /**
     * Custom implementation for Traveling Salesman Problem. It's not possible
     * to do a standard crossover, because we can't duplicate cityIDs (part 1 of
     * parent 1 would most likely contain duplicates of some part of parent 2).
     * 
     * Here is the basic 3-step process: 1. copy some cityIDs from left side of
     * chromosome1 (same as original algorithm) 2. copy non-duplicate cityIDs
     * from rigt side of chromosome2 3. copy a randomized list of the remaining
     * cityIDs
     * 
     * To avoid duplicates during step 2 and 3, keep a checklist of which
     * cityIDs have been copied so far. The helper class IntegerChecklist will
     * keep track of these and return the cityIDs not used. The final step is to
     * randomize these remaining cityIDs, because they will be returned from
     * IntegerChecklist in numerical order (which is not a good feature in a
     * Genetic Algorithm).
     * 
     * A note on implementation and naming: I used vague names like
     * "currentValue" and "remainingValues" because, although this code is being
     * used for the TSP in the context of cityIDs, it could be used for any
     * linear, unique set of int's.
     */
    public Chromosome singlePointCrossover(Chromosome chromosome1, Chromosome chromosome2) {
        int crossoverPoint = (int) (random.nextDouble() * allelesPerChromosome);
        if (random.nextDouble() >= crossoverProbability) {
            // If no crossover then randomly return one parent
            return random.nextBoolean() ? chromosome1.deepCopy() : chromosome2.deepCopy();
        }

        // here, we don't need to initialize a TSP-specific chromosome (by
        // calling ProblemTSP.getTemplateChromosome()), because
        // we are going to overwrite all the alleles anyway.
        Chromosome result = new IntChromosome(new int[allelesPerChromosome]);

        // Copying the first part is the same as the basic algorithm
        // NOTE: This step preserves the Zero-City Anchor constraint (that index
        // 0 must be cityID 0). If this method is changed at all, it will be
        // necessary to preserve this constraint somehow.
        result.partiallyCopyFromOtherChromosome(chromosome1, 0, crossoverPoint - 1);

        // check off values that have been copied from chromosome1
        IntegerChecklist checklist = new IntegerChecklist(allelesPerChromosome);
        for (int i = 0; i < crossoverPoint; i++) {
            checklist.checkOff(((IntChromosome) result).getAllele(i));
        }

        // loop through parent 2 and add what is not checked off yet
        // (and check it off)
        int childIndex = crossoverPoint;
        for (int parentIndex = crossoverPoint; parentIndex < allelesPerChromosome; parentIndex++) {
            int currentInteger = ((IntChromosome) chromosome2).getAllele(parentIndex);
            if (!checklist.isCheckedOff(currentInteger)) {
                ((IntChromosome) result).setAllele(childIndex, currentInteger);
                checklist.checkOff(currentInteger);
                childIndex++;
            }
        }

        // get the fallout list - what integers haven't been added yet?
        int[] remainingValues = checklist.getValuesNotYetCheckedOff();
        // randomize the remaining values
        RandomIntSequence randomIntSequence = new RandomIntSequence(remainingValues);
        // add the randomized list to the remaining empty slots on the result
        while (childIndex < allelesPerChromosome) {
            ((IntChromosome) result).setAllele(childIndex++, randomIntSequence.getNext());
        }

        return result;
    }
}
